import re
from pathlib import Path
from collections.abc import Iterable
from io import BytesIO
from fontTools.ttLib import TTFont

from PIL import Image, ImageDraw, ImageFont, ImageFilter
from PIL.ImageFont import FreeTypeFont
from PIL.Image import Image as IMG

import matplotlib.font_manager as fm


class FontManager:
    def __init__(self, font_name: str, fallback: list[str], size: Iterable[int] | None = None) -> None:
        self.font_name: str = font_name
        self.fallback: list[str] = fallback
        self.cmap = TTFont(fm.findfont(fm.FontProperties(family=font_name)), fontNumber=0).getBestCmap()
        self.fallback_cmap = {}
        for font in self.fallback:
            font_path = fm.findfont(fm.FontProperties(family=font))
            self.fallback_cmap[font_path] = TTFont(font_path, fontNumber=0).getBestCmap()

        self.font_def = {k: self.new_font(font_name, k) for k in size} if size else {}

    def font(self, size: int):
        return self.font_def.get(size) or self.new_font(self.font_name, size)

    def new_font(self, name: str, size: int):
        return ImageFont.truetype(font=name, size=size, encoding="utf-8")


def linecard_to_png(text: str, font_manager: FontManager, **kwargs):
    """
    文字转png
    """
    output = BytesIO()
    linecard(text, font_manager, **kwargs).save(output, format="png")
    return output


def remove_tag(text: str, pattern: re.Pattern):
    match = pattern.search(text)
    if match:
        start = match.start()
        end = match.end()
        return text[:start] + text[end:], text[start:end]
    else:
        return None


def line_wrap(line: str, width: int, font: FreeTypeFont, start: float = 0):
    text_x = start
    new_str = ""
    for char in line:
        text_x += font.getlength(char)
        if text_x > width:
            new_str += "\n" + char
            text_x = 0
        else:
            new_str += char
    return new_str


class Tag:
    def __init__(self, font, cmap, align="left") -> None:
        self.align: str = align
        self.font: FreeTypeFont = font
        self.cmap: dict = cmap
        self.color: str | None = None
        self.passport: bool = False
        self.noautowrap: bool = False
        self.nowrap: bool = False


class linecard_pattern:
    align = re.compile(r"\[left\]|\[right\]|\[center\]|\[pixel\]\[.*?\]")
    font = re.compile(r"\[font\]\[.*?\]\[.*?\]")
    color = re.compile(r"\[color\]\[.*?\]")
    passport = re.compile(r"\[passport\]")
    nowrap = re.compile(r"\[nowrap\]")
    noautowrap = re.compile(r"\[noautowrap\]")


def linecard(
    text: str,
    font_manager: FontManager,
    font_size: int,
    width: int | None = None,
    height: int | None = None,
    padding: tuple[int, int] = (20, 20),
    spacing: float = 1.2,
    bg_color: str | int = 0,
    autowrap: bool = False,
    canvas: IMG | None = None,
) -> IMG:
    """
    指定宽度单行文字
        ----:横线

        [left]靠左
        [right]靠右
        [center]居中
        [pixel][400]指定像素

        [font][font_name][font_size]指定字体

        [color][#000000]指定本行颜色

        [nowrap]不换行
        [noautowrap]不自动换行
        [passport]保持标记
    """
    text = text.replace("\r\n", "\n")
    lines = text.split("\n")
    padding_x, padding_y = padding

    align = "left"

    font_def = font_manager.font(font_size)
    cmap_def = font_manager.cmap

    tag = Tag(font_def, cmap_def)

    x, max_x, y, charlist = (0.0, 0.0, 0.0, [])
    for line in lines:
        # 检查继承格式
        if tag.passport:
            tag.passport = False
        else:
            tag.__init__(font_def, cmap_def, align="nowrap" if tag.nowrap else "left")

        # 检查对齐格式
        if data := remove_tag(line, linecard_pattern.align):
            line, align = data
            if align.startswith("[pixel]["):
                tag.align = align[8:-1]
                x = 0
            else:
                tag.align = align[1:-1]

        if data := remove_tag(line, linecard_pattern.font):
            line, font = data
            if font.startswith("[font]["):
                font = font[7:-1]
                inner_font_name, inner_font_size = font.split("][", 1)
                inner_font_size = int(inner_font_size) if inner_font_size else font_size
                inner_font_name = inner_font_name or font_def.path
                try:
                    tag.font = ImageFont.truetype(font=inner_font_name, size=inner_font_size, encoding="utf-8")
                    tag.cmap = TTFont(tag.font.path, fontNumber=tag.font.index).getBestCmap()
                except OSError:
                    pass

        if data := remove_tag(line, linecard_pattern.color):
            line, color = data
            tag.color = color[8:-1]

        if data := remove_tag(line, linecard_pattern.noautowrap):
            line = data[0]
            tag.noautowrap = True

        if data := remove_tag(line, linecard_pattern.nowrap):
            line = data[0]
            tag.nowrap = True
        else:
            tag.nowrap = False

        if data := remove_tag(line, linecard_pattern.passport):
            line = data[0]
            tag.passport = True

        if autowrap and not tag.noautowrap and width and tag.font.getlength(line) > width:
            line = line_wrap(line, width - sum(padding), tag.font, x)

        if line == "----":
            inner_tmp = tag.font.size * spacing
            charlist.append([line, None, y, inner_tmp, tag.color, None])
            y += inner_tmp
        else:
            line_segs = line.split("\n")
            for seg in line_segs:
                for char in seg:
                    ord_char = ord(char)
                    inner_font = tag.font
                    if ord_char not in tag.cmap:
                        for (
                            fallback_font,
                            fallback_cmap,
                        ) in font_manager.fallback_cmap.items():
                            if ord_char in fallback_cmap:
                                inner_font = ImageFont.truetype(
                                    font=fallback_font,
                                    size=tag.font.size,
                                    encoding="utf-8",
                                )
                                break
                        else:
                            char = "□"
                    charlist.append([char, x, y, inner_font, tag.color, tag.align])
                    x += inner_font.getlength(char)
                max_x = max(max_x, x)
                x, y = (x, y) if tag.nowrap else (0, y + tag.font.size * spacing)

    width = width if width else int(max_x + padding_x * 2)
    height = height if height else int(y + padding_y * 2)
    canvas = canvas if canvas else Image.new("RGBA", (width, height), bg_color)
    draw = ImageDraw.Draw(canvas)

    for i, (char, x, y, font, color, align) in enumerate(charlist):
        if char == "----":
            color = color if color else "gray"
            inner_y = y + (font - 0.5) // 2 + padding_y
            draw.line(((0, inner_y), (width, inner_y)), fill=color, width=4)
        else:
            if align == "left":
                start_x = padding_x
            elif align == "nowrap":
                pass
            elif align.isdigit():
                start_x = int(align)
            else:
                for inner_i, inner_y in enumerate(map(lambda x: (x[2]), charlist[i:])):
                    if inner_y != y:
                        inner_index = charlist[i + inner_i - 1]
                        break
                else:
                    inner_index = charlist[-1]
                inner_char = inner_index[0]
                inner_font = inner_index[3]
                inner_x = inner_index[1]
                inner_x += inner_font.getlength(inner_char)
                if align == "right":
                    start_x = width - inner_x - padding_x
                elif align == "center":
                    start_x = (width - inner_x) // 2
                else:
                    start_x = padding_x
            color = color if color else "black"
            draw.text((start_x + x, y + padding_y), char, fill=color, font=font)

    return canvas


ImageList = list[IMG]


def info_splicing(
    info: ImageList,
    BG_path: Path | None = None,
    width: int = 880,
    padding: int = 20,
    spacing: int = 20,
    BG_type: str = "GAUSS",
):
    """
    信息拼接
        info:信息图片列表
        bg_path:背景地址
    """

    height = padding
    for image in info:
        # x = image.size[0] if x < image.size[0] else x
        height += image.size[1]
        height += spacing * 2
    else:
        height = height - spacing + padding

    size = (width + padding * 2, height)
    if BG_path and BG_path.exists():
        bg = Image.open(BG_path).convert("RGB")
        canvas = CropResize(bg, size)
    else:
        canvas = Image.new("RGB", size, "white")
        BG_type = "NONE"

    height = padding

    def colorBG(canvas: IMG, image: IMG):
        colorBG = Image.new("RGBA", (width, image.size[1]), BG_type)
        canvas.paste(colorBG, (20, height), mask=colorBG)
        canvas.paste(image, (20, height), mask=image)

    def blurBG(canvas: IMG, image: IMG):
        box = (20, height, 900, height + image.size[1])
        region = canvas.crop(box)
        blurred_region = region.filter(ImageFilter.GaussianBlur(radius=4))
        canvas.paste(blurred_region, box)
        canvas.paste(image, (20, height), mask=image)

    def noneBG(canvas: IMG, image: IMG):
        canvas.paste(image, (20, height), mask=image)

    funcBG = {"GAUSS": blurBG, "NONE": noneBG}.get(BG_type, colorBG)

    for image in info:
        funcBG(canvas, image)
        height += image.size[1]
        height += spacing * 2
    output = BytesIO()
    canvas.convert("RGB").save(output, format="png")
    return output


def CropResize(img: IMG, size: tuple[int, int]):
    """
    修改图像尺寸
    """

    test_x = img.size[0] / size[0]
    test_y = img.size[1] / size[1]

    if test_x < test_y:
        width = img.size[0]
        height = size[1] * test_x
    else:
        width = size[0] * test_y
        height = img.size[1]

    center = (img.size[0] / 2, img.size[1] / 2)
    output = img.crop(
        (
            int(center[0] - width / 2),
            int(center[1] - height / 2),
            int(center[0] + width / 2),
            int(center[1] + height / 2),
        )
    )
    output = output.resize(size)
    return output



import importlib
import time
import re
from pathlib import Path
from collections.abc import Callable, Coroutine, Iterable, Sequence
from .logger import logger


class PluginError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class Result:
    def __init__(self, send_method: str, data) -> None:
        self.send_method = send_method
        self.data = data


class Event:
    def __init__(
        self,
        raw_command: str,
        args: Sequence[str],
    ):
        self.raw_command = raw_command
        self.args = args
        self.kwargs: dict = {}
        self.get_kwargs: dict[str, Callable[..., Coroutine]] = {}


class Handle:
    func: Callable[[Event], Coroutine[None, None, Result | None]]

    def __init__(self, extra_args: Iterable[str], get_extra_args: Iterable[str]):
        self.extra_args = extra_args
        self.get_extra_args = get_extra_args

    async def __call__(self, event: Event):
        return await self.func(event)


PluginCommands = str | set[str] | re.Pattern | None


class Plugin:
    def __init__(
        self,
        name: str = "",
        build_event=None,
        build_result=None,
    ) -> None:
        self.name: str = name
        self.handles: dict[int, Handle] = {}
        self.command_dict: dict[str, set[int]] = {}
        self.regex_dict: dict[re.Pattern, set[int]] = {}
        self.temp_handles: dict[str, tuple[float, Handle]] = {}
        self.startup_tasklist: list[Coroutine] = []
        self.shutdown_tasklist: list[Coroutine] = []
        self.build_event: Callable | None = build_event
        self.build_result: Callable | None = build_result

    def handle_warpper(self, func: Callable[..., Coroutine]):
        if build_event := self.build_event:
            middle_func = lambda e: func(build_event(e))
        else:
            middle_func = func

        if build_result := self.build_result:
            async def wrapper(event):
                if result := await middle_func(event):
                    return build_result(result)

            return wrapper
        else:
            return middle_func

    def commands_register(self, commands: PluginCommands, key: int):
        if not commands:
            self.command_dict.setdefault("", set()).add(key)
        elif isinstance(commands, set):
            for command in commands:
                self.command_dict.setdefault(command, set()).add(key)
        elif isinstance(commands, str):
            self.regex_dict.setdefault(re.compile(commands), set()).add(key)
        elif isinstance(commands, re.Pattern):
            self.regex_dict.setdefault(commands, set()).add(key)
        else:
            raise PluginError(f"指令：{commands} 类型错误：{type(commands)}")

    def handle(
        self,
        commands: PluginCommands,
        extra_args: Iterable[str] = [],
        get_extra_args: Iterable[str] = [],
    ):
        def decorator(func: Callable[..., Coroutine]):
            key = len(self.handles)
            self.commands_register(commands, key)
            handle = Handle(extra_args, get_extra_args)
            handle.func = self.handle_warpper(func)
            self.handles[key] = handle

        return decorator

    def temp_handle(
        self,
        key: str,
        extra_args: Iterable[str] = [],
        get_extra_args: Iterable[str] = [],
        timeout: float | int = 30.0,
    ):

        def decorator(func: Callable[..., Coroutine]):
            def finish():
                del self.temp_handles[key]

            handle = Handle(extra_args, get_extra_args)
            handle.func = self.handle_warpper(lambda e: func(e, finish))
            self.temp_handles[key] = time.time() + timeout, handle

        return decorator

    def startup(self, func: Callable[[], Coroutine]):
        """注册一个启动任务"""
        self.startup_tasklist.append(func())

        return func

    def shutdown(self, func: Callable[[], Coroutine]):
        """注册一个结束任务"""
        self.shutdown_tasklist.append(func())

        return func

    def __call__(self, command: str) -> dict[int, Event] | None:
        command_list = command.split()
        if not command_list:
            return
        data = {}
        command_start = command_list[0]
        for cmd, keys in self.command_dict.items():
            if not command_start.startswith(cmd):
                continue
            if command_start == cmd:
                args = command_list[1:]
            else:
                command_list[0] = command_list[0][len(cmd) :]
                args = command_list
            event = Event(command, args)
            data.update({key: event for key in keys})

        for pattern, keys in self.regex_dict.items():
            if args := re.match(pattern, command):
                event = Event(command, args.groups())
                data.update({key: event for key in keys})

        return data

    def temp_check(self) -> bool:
        if not self.temp_handles:
            return False
        now = time.time()
        self.temp_handles = {k: v for k, v in self.temp_handles.items() if v[0] > now}
        if not self.temp_handles:
            return False
        return True


class PluginLoader:
    def __init__(self, plugins_path: str | Path | None = None, plugins_list: list[str] = []) -> None:
        self.plugins_path: Path | None = None if plugins_path is None else Path(plugins_path)
        self.plugins_list: list = plugins_list

    @staticmethod
    def load(name: str) -> Plugin | None:
        logger.info(f"【loading plugin】 {name} ...")
        try:
            module = importlib.import_module(name)
            plugin = getattr(module, "__plugin__", None)
            if not isinstance(plugin, Plugin):
                return
            plugin.name = plugin.name or name
            return plugin
        except:
            logger.exception(name)

    def plugins_from_path(self):
        if self.plugins_path is None:
            return []
        plugins_path = ".".join(self.plugins_path.relative_to(Path()).parts)
        plugins = []
        for x in self.plugins_path.iterdir():
            name = x.stem if x.is_file() and x.name.endswith(".py") else x.name
            if name.startswith("_"):
                continue
            plugins.append(self.load(f"{plugins_path}.{name}"))
        return [plugin for plugin in plugins if plugin]

    def plugins_from_list(self):
        return [plugin for x in self.plugins_list if (plugin := self.load(x))]

    @property
    def plugins(self) -> list[Plugin]:
        return self.plugins_from_path() + self.plugins_from_list()



from collections.abc import Iterable
from typing import overload


class Library[K, V]:
    """多键字典"""

    @overload
    def __init__(self, data: Iterable[tuple[K, Iterable[K], V]]) -> None: ...
    @overload
    def __init__(self) -> None: ...

    def __init__(self, data=None) -> None:
        self._key_data: dict[K, V] = {}
        self._index_key: dict[K, K] = {}
        self._key_indices: dict[K, set[K]] = {}
        if not data:
            return
        for key, alias, value in data:
            self.set_item(key, alias, value)

    def __getitem__(self, index: K) -> V:
        return self._key_data.get(index) or self._key_data[self._index_key[index]]

    def __setitem__(self, index: K, data: V):
        key = self._index_key.get(index, index)
        self._key_data[key] = data

    def __contains__(self, key):
        return key in self._key_indices or key in self._index_key

    def __iter__(self):
        return iter((key, self._key_indices[key], value) for key, value in self._key_data.items())

    def keys(self):
        return self._key_data.keys()

    def values(self):
        return self._key_data.values()

    def items(self):
        return self._key_data.items()

    def set_item(self, key: K, indices: Iterable[K], data: V):
        if old_indices := self._key_indices.get(key):
            for i in old_indices:
                del self._index_key[i]
        self._key_data[key] = data
        indices = set(indices)
        self._key_indices[key] = indices
        for i in indices:
            self._index_key[i] = key

    def __delitem__(self, index: K):
        if index in self._key_data:
            del self._key_data[index]
            if indices := self._key_indices.get(index):
                for i in indices:
                    del self._index_key[i]
                del self._key_indices[index]
            return
        if key := self._index_key[index]:
            del self._index_key[index]
            self._key_indices[key].discard(index)
            return
        raise KeyError(index)

    def update(self, data: "Library[K, V]"):
        for key, indices, value in data:
            self.set_item(key, indices, value)

    @overload
    def get(self, index: K) -> V | None: ...
    @overload
    def get(self, index: K, default: V) -> V: ...

    def get(self, index: K, default=None):
        if key := self._index_key.get(index):
            return self._key_data[key]
        return self._key_data.get(index, default)

    def setdefault(self, index: K, default: V):
        if key := self._index_key.get(index):
            return self._key_data[key]
        return self._key_data.setdefault(index, default)



import asyncio
from collections.abc import Awaitable
from .core.plugin import Plugin, PluginLoader, Event
from .core.adapter import Adapter
from .core.logger import logger


class Clovers:
    def __init__(self) -> None:
        self.global_adapter: Adapter = Adapter()
        self.plugins: list[Plugin] = []
        self.adapter_dict: dict[str, Adapter] = {}
        self.plugins_dict: dict[str, list[Plugin]] = {}
        self.wait_for: list[Awaitable] = []

    async def response(self, adapter_key: str, command: str, **extra) -> int:
        task_list = []
        adapter = self.adapter_dict[adapter_key]
        plugins = self.plugins_dict[adapter_key]
        for plugin in plugins:
            if data := plugin(command):
                task_list.extend(adapter.response_safe(plugin.handles[key], event, extra) for key, event in data.items())
            if plugin.temp_check():
                event = Event(command, [])
                task_list.extend(adapter.response_safe(handle, event, extra) for _, handle in plugin.temp_handles.values())
        return sum(await asyncio.gather(*task_list)) if task_list else 0

    def load_plugin(self, name: str):
        if self.wait_for:
            raise RuntimeError("cannot loading plugin after clovers startup")
        plugin = PluginLoader.load(name)
        if plugin is None:
            logger.error(f"未找到 {name}")
        elif plugin not in self.plugins:
            self.plugins.append(plugin)
            logger.info(f"{name} 加载成功")
        else:
            logger.info(f"{name} 已存在")

    async def startup(self):
        self.wait_for.extend(asyncio.create_task(task) for plugin in self.plugins for task in plugin.startup_tasklist)
        self.wait_for.extend(task for plugin in self.plugins for task in plugin.shutdown_tasklist)
        # 混合全局方法
        # 过滤没有指令响应任务的插件
        # 检查任务需求的参数是否存在于响应器获取参数方法。
        extra_args_dict: dict[str, set[str]] = {}
        for adapter_key, adapter in self.adapter_dict.items():
            adapter.remix(self.global_adapter)
            extra_args_dict[adapter_key] = set(adapter.kwarg_dict.keys())
            self.plugins_dict[adapter_key] = []

        for plugin in self.plugins:
            if not plugin.handles:
                continue
            extra_args: set[str] = set()
            extra_args = extra_args.union(*[set(handle.extra_args) | set(handle.get_extra_args) for handle in plugin.handles.values()])
            for adapter_key, existing in extra_args_dict.items():
                if method_miss := extra_args - existing:
                    logger.warning(
                        f'插件 "{plugin.name}" 声明了适配器 "{adapter_key}" 未定义的kwarg方法',
                        extra={"method_miss": method_miss},
                    )
                    logger.debug(f'"{adapter_key}"未定义的kwarg方法:{method_miss}')
                else:
                    self.plugins_dict[adapter_key].append(plugin)
        self.plugins.clear()

    async def shutdown(self):
        await asyncio.gather(*self.wait_for)

    async def __aenter__(self) -> None:
        await self.startup()

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.shutdown()
